"""SPSSbench module for timing command sets
Note that it requires the Pywin32 extensions, which can be found here
http://sourceforge.net/projects/pywin32/

Copyright (C) 2005 by SPSS Inc.

Usage example within SPSS:
begin program.
import spss
import benchmark

spss.Submit("get file='c:/spss14/1991 U.S. General Social Survey.sav'.")
bstat = benchmark.benchstats(\
["WorkingSetSize","PeakWorkingSetSize", "ReadOperationCount",
"WriteOperationCount","UserTime","KernelTime"] )

cmdlist1='''FREQ happy childs educ.'''
cmdlist2='''DESC happy childs educ.'''

benchmark.benchmark(numrep=10, cmdlist1=cmdlist1,  bstat=bstat, file="c:/temp/gbench.txt")
end program.
GET DATA  /TYPE = TXT  /FILE = 'C:\temp\gbench.txt'
 /DELCASE = LINE /DELIMITERS = " " /ARRANGEMENT = DELIMITED /FIRSTCASE = 2 /IMPORTCASE = ALL
 /VARIABLES =
 Group F1.0
 timeStart F18.2
 timeStop F9.2
 WorkingSetSizeStart F9.2
 WorkingSetSizeStop F5.2
 PeakWorkingSetSizeStart F5.2
 PeakWorkingSetSizeStop F9.2
 ReadOperationCountStart F9.2
 ReadOperationCountStop F18.2
 WriteOperationCountStart F9.2
 WriteOperationCountStop F9.2
 UserTimeStart F5.2
 UserTimeStop F5.2
 KernelTimeStart F9.2
 KernelTimeStop F9.2.
"""

__all__ = [
    "benchstats",
    "benchmark",
    ]

__author__ =  'spss'
__version__=  '1.0'

import spss
import time, re
try:
    from win32process import GetCurrentProcess, GetProcessMemoryInfo, GetProcessTimes, GetProcessIoCounters
except:
    print "This module requires win32process, which can be downloaded from http://sourceforge.net/projects/pywin32/"
    raise

class benchstats(object):
    """This class handles benchmark measures and snapshots.  It is Windows specific.
    It monitors the current process (expected to be SPSS), so any overhead from the
    monitoring will be included in the measures.  By default, the measures collected
    are just elapsed times.
    """
    memory = [
    'QuotaPagedPoolUsage', 'QuotaPeakPagedPoolUsage', 'QuotaNonPagedPoolUsage',
    'PageFaultCount', 'PeakWorkingSetSize', 'PeakPagefileUsage', 'QuotaPeakNonPagedPoolUsage',
    'PagefileUsage', 'WorkingSetSize']
    time = [
    'CreationTime', 'UserTime', 'KernelTime']
    io = [
            'WriteOperationCount','WriteTransferCount', 'OtherOperationCount',
    'OtherTransferCount', 'ReadOperationCount', 'ReadTransferCount']
    apinames = [GetProcessMemoryInfo, GetProcessTimes,GetProcessIoCounters]
    apilist = [memory, time, io]

    def __init__(self, stats=[]):
        if not isinstance(stats, list):    #type(stats) is not list:
            stats = [stats]
        self.stats = stats
        self.readings = []
        self.measures=["time"]
        self.apis = [time]
        self.process = GetCurrentProcess()
        
# build list of calls for specified measures
        for s in stats:
            for i, api in enumerate(benchstats.apilist):
                if s in api:
                    self.measures.append(s)
                    self.apis.append(benchstats.apinames[i])
                    break
            else:
                raise ValueError, "Invalid measure: " + s

    def snap(self, group=1):
        "record new set of statistics measures for designated group, which defaults to 1"
        r = [group]
        for i in range(len(self.measures)):
            if self.measures[i] == "time":
                r.append(time.time())
            else:
                r.append(self.apis[i](self.process)[self.measures[i]])
        self.readings.append(tuple(r))

    def getrows(self, mergecount=1):
        """generator for sets of snap measures.  It yields the next complete row of snaps.

        mergecount specifies the number of
        snaps to be combined to make a complete observation.  If snapping before and after some
        events, for example, mergecount should be 2."""

        assert mergecount, "Number of rows to merge must be positive"

        for rows in range(0, len(self.readings), mergecount):
            row = str(self.readings[rows][0])+ " " + str([self.readings[rows+i][1:] for i in range(mergecount)])
            yield re.sub(r"[][(),L]", "", row) + "\n"

    def save(self, filespec, sets=[""]):
        """write measures to specified file in a form suitable for reading into SPSS.
        sets is a list of strings to append to the generated variable names.  Its length
        determines how readings are paired.  If the length is two, for example, there will
        be 2 times the number of statistics in each row with the names suffixed by the
        strings given.  The size of sets determines how many (sequential) snaps are
        concatenated to give each output record.

        For example:
        bstat.save(file, sets=["Start", "Stop"])
        would save a set of before and after measures for the selected statistics with
        Start and Stop appended to the pairs of variable names.
        """

        if not isinstance(sets, list):      #type(sets) is not list:
            sets = [sets]
        f = open(filespec, "w")
#   construct variable names heading
        namelist = ["Group"]
        for s in sets:
            for i in range(len(self.measures)):
                namelist.append(self.measures[i] + s)   
        f.write(" ".join(namelist)+"\n")

        for row in self.getrows(mergecount=len(sets)):
            f.write(row)
        f.close()



def benchmark(numrep, bstat, cmdlist1, cmdlist2=None,file=None):
    """Benchmark repetitions of one or more commands against, optionally, an alternative set.
    When there are two sets, the repetitions are interleaved.
    numrep is the repetition count.
    cmdlist1 and cmdlist 2 are lists of commands to be timed.  Remember to include
    an EXECUTE if you are timing transformation commands that would not otherwise
    generate a data pass.
    bstat, required, is a benchstats object containing the statistics to be collected.
    
    File, is specified, gets the individual results for the specified measures in a form
    that is is suitable for reading into SPSS.  If file is not specified, all the commands
    will be run, but there will be no measures written
    Variable names are written on the first line.
    """
    numsets = cmdlist2 and 2 or 1
    
    for i in range(numrep):
        bstat.snap(group=1)     #start
        spss.Submit(cmdlist1)
        bstat.snap(group=1)     #stop
        if cmdlist2:
            bstat.snap(group=2) #start
            spss.Submit(cmdlist2)
            bstat.snap(group=2) #stop

    if file:
        bstat.save(file, sets=["Start", "Stop"])
