"""A collection of special transformations extending those built in to SPSS"""
# This module provides special transformations.  It will grow as new types of transformations are identified.
# Current contents include
# Reversecode: use the value labels from one variable as a mapping from labels to values for another variable whose values are strings.
# (The more obvious mapping, from values to labels can be calculated using the VALUELABEL function in SPSS.)

# Recode: use an arbitrary dictionary of keys and values to recode a variable into a new variable and label the values.  This is a
# convenience repackaging of some of the functionality of the SPSS RECODE command plus automatic value labels and
# recognition of date literals on input.

# CollapseInfrequentValues: combine values with low occurrence into an "other" category



#Copyright (C) 2007 by SPSS Inc.

# history
# 19-jun-2006   jkp   Initial version
# 30-nov-2006  jkp   Added extendedNtiles transformation
# 23-apr-2007  jkp   Added collapseInfrequentValues function
# 26-jun-2007  jkp   Add rank-based collaping


__version__ = '1.3.0'
__author__ = 'spss'


import spss, spssaux, spssdata, re, operator

class Reversecode(object):
    def __init__(self, newvar, inputvar, valuelabels, format="F8.0"):
        """Create a numeric variable whose values are the values labeled by a value label set, and create
        corresponding value labels for this variable.
        newvar is a string containing the name for the output variable.  It must not already exist in the dataset.
        inputvar is a string containing the name of the variable to be transformed.  It must be a string variable.
        valuelabel is a dictionary of values and valuelabels that will define the mapping from labels to values.  It
        can be created most easily from the spssaux.VariableDict method.
        format is an SPSS format to be applied to the new variable with a default of F8.0
        
        If a value is encountered in the transformation that is not defined in the value label mapping, the result will be missing.
        Both the label and string values are right-trimmed before testing for equality."""
        
        self.newvar = newvar
        self.inputvar = inputvar
        self.valuelabels = valuelabels
        self.format = format
        
    def generate(self,execute=True):
        """Generate the transformation.  If execute is True, the default, the syntax is generated and run.
        The method returns the generated syntax as a list of two commands, one for the transformation and one for the value labels."""

        exprlist = []
        labellist = []
        for  val, label in self.valuelabels.items():
            exprlist.append("RTRIM("+self.inputvar + ")=RTRIM(" + _smartquote(label) + ").\nCOMPUTE "+ self.newvar + "=" + val + ".")
        cmd = "DO IF " + "\nELSE IF ".join(exprlist) + "\nEND IF."
        vl = [str(val) + " " + _smartquote(label) for val, label in self.valuelabels.items()]
        vlcmd = "VALUE LABELS " + self.newvar + " " + " ".join(vl) + "."
        if execute:
            spss.Submit(["NUMERIC " + self.newvar + "("+self.format+").", cmd, vlcmd])
        return [cmd, vlcmd]
    
def ReversecodeComplete(newvar, inputvar,vallabelsvar):
    "compute a new variable, newvar, from the variable inputvar using the value labels from variable vallabels."
    
    vardict = spssaux.VariableDict(vallabelsvar)
    code = Reversecode(newvar, inputvar, vardict[vallabelsvar].ValueLabels)
    gencode = code.generate(execute=True)
    return gencode
        
class Recode(object):
    """Map the values of a variable into a new variable according to a supplied dictionary.  This class uses the SPSS
    RECODE command but adds support for date literals in the input and generates value labels for the output variable."""
    
    def __init__(self, newvar, inputvar,  mapping, inputtype="F", format="F8.0", elserc=None, makevaluelabels=True):
        """newvar is a string naming a new variable.
        inputvar is a string naming an existing variable.
        inputtype is the type of the input variable, which must be "F", "A", or "D" for numeric, string, and (numeric) date.
        mapping is a dictionary that defines the mapping.  A key may be a simple value of the appropriate type or
        a tuple representing a range.  (Note that ranges are not valid for strings in the SPSS Recode command.)
        The mapping requests are ordered by the output value, and the first specification covering an input value determines
        how RECODE will map the value.  This matters only if the input values overlap.
        If inputtype is D and a key value contains a string, it is interpreted as a date in the format yyyy-mm-dd.
        format is the format for the new variable, defaulting to F8.0.  Format letter must be F or A.
        if valuelabels is True, the default, the mapping is used to generate value labels for the new variable.
        For many-to-one mappings, the labels are merged.
        Any values encountered in inputvar not defined in mapping will, by default, be mapped to sysmis (if numeric) and not labeled.
        If elserc=value is specified, that value will be used for any other values.  If elserc is an empty tuple, unmapped values are
        copied.
        Examples:
        rec = specialtransforms.Recode("educrecode", "educ",{8:1, 12:2, 16:3, 17:3})
        syn = rec.generate(execute=False)

        rec = specialtransforms.Recode("educrecode", "educ",{(8,11):1,(18,"HIGH"):4, 12:2, 16:3, 17:3})
        syn = rec.generate(execute=False)
        """
        
        self.newvar = newvar
        self.inputvar = inputvar
        self.mapping = mapping
        self.format = format
        if not inputtype.upper() in ["F", "A", "D"]:
            raise ValueError, "inputtype must be F, A, or D"
        if not format.upper()[0] in ["F", "A"]:
            raise ValueError, "format must be of F or A type"
        self.inputtype = inputtype
        self.makevaluelabels = makevaluelabels
        self.elserc = elserc
        
    def generate(self, execute=True):
        """generate the recode and valuelabel syntax, which is returned as a two-element list.
        If execute is True, the default, execute the transformations."""

        if self.format.upper()[0] == "A" and self.format[1].isdigit():
            typecmd = "STRING"
            outquo = '"'
        else:
            typecmd = "NUMERIC"
            outquo = ""
        typecmd = typecmd + " " + self.newvar + " (" + self.format + ")."
        recspec = _recodespec(self.mapping, self.inputtype, self.format[0])
        recodelist = ["(" + inspec + "=" + outspec + ")" for (inspec, outspec) in recspec.recodelist()]
        if self.elserc is None:
            elsetext = ""
        elif type(self.elserc) is tuple:
           elsetext = " (ELSE=COPY)"
        else:
            elsetext = " (ELSE=" + _smartquote(self.elserc, outquo) + ")"
        recodecmd = "RECODE " + self.inputvar + " " + " ".join(recodelist) + elsetext + " INTO " + self.newvar + "."

        if self.makevaluelabels and (outquo == "" or int("".join([d for d in list(self.format) if d.isdigit()])) <=8):    # not a long string (restriction to be removed eventually)
            vllist = [_smartquote(key, outquo) + " " + _smartquote(val) for key, val in  recspec.vllist.items()]
        else:
            vllist = None

        if execute:
            spss.Submit([typecmd, recodecmd])
        if vllist:
            vlcmd = "VALUE LABELS " + self.newvar + " " + " ".join(vllist) + "."
            if execute:
                spss.Submit(vlcmd)
        else:
            vlcmd = ""
        return [typecmd, recodecmd, vlcmd]
    
class _recodespec(object):
    """Construct and maintain a recode list including both points and intervals"""
    def __init__(self, recodedic, inputtype, outputtype):
        """recodedic is a dictionary mapping input values, keys, to outputs, values.  Each key may be a single value or a duple.
        The recodespec is ordered by the output code, which determines the priority in case of overlapping codes.
        If the inputtype is "D", string values are converted to date numerics."""
        self.vllist = _makevaluelabels(recodedic)
        self.rlist = recodedic.items()
        self.rlist.sort(key=lambda x:x[-1])    # sort by output value
        self.inputtype = inputtype
        self.outputtype = outputtype
        
    def recodelist(self):
        """Yield recode specs in appropriate form for RECODE command"""
        for item in self.rlist:
            input, output = item[0], item[1]
            if not isinstance(input, tuple):
                input = valuespec(input, self.inputtype)
            else:
                input = valuespec(input[0], self.inputtype) + " THRU " + valuespec(input[1], self.inputtype)
            yield (input, valuespec(output, self.outputtype))
            
def valuespec(value, valuetype):
    """format value appropriately.  Quote if string; convert date values to numbers if needed"""
    if isinstance(value, basestring):
        if valuetype != "A" and value.upper() in ['LOW', 'LO', 'HIGH', 'HI']:
            return value
        if valuetype == "D":
            return str(yrmodasec(value.split("-")))
        else:
            return _smartquote(value)
    else:
        return str(value)
            

def _makevaluelabels(mapping):
    """merge value labels in case of many-to-one mapping"""
    d = {}
    killlist = ["LOW", "LO", "HIGH", "HI"]
    for key, value in mapping.items():
        if isinstance(key, tuple):
            key = list(key)
            for i in [0,1]:
                if str(key[i]) in killlist:
                       key[i] = " "
            keystr = "("+str(key[0]) + " - " + str(key[1]) + ")"
        else:
            keystr = str(key)
        if d.has_key(value):
            d[value] = d[value] + ", " + keystr
        else:
            d[value] = keystr
    return d

def _smartquote(s, qchar='"'):
    """ smartquote a string so that internal quotes are distinguished from surrounding
    quotes for SPSS and return that string with the surrounding quotes.  qchar is the
    character to use for surrounding quotes."""
    return qchar + str(s).replace(qchar, qchar+qchar) + qchar
            

def yrmodasec(ymd):
    """compute SPSS internal date value from four digit year, month, and day.
    ymd is a list of numbers in that order.  The parts will be truncated to integers.
    The result is equivalent to the SPSS subroutine yrmoda result converted to seconds"""
    
    if len(ymd) != 3:
        raise ValueError, "date specification must have the form yyyy-mm-dd"
    year = int(ymd[0])
    month = int(ymd[1])
    day = int(ymd[2])
    
    if year < 1582 or month < 1 or month > 13 or day <0 or day > 31:
        raise ValueError, ("Invalid date value: %d %d %d") % (year, month, day)
    yrmo = year * 365 + (year+3)//4 - (year+99)//100 + (year + 399)//400 \
         + 3055 *(month+2)//100 - 578192
    if month > 2:
        yrmo-= 2
        if (year%4 == 0 and (year%100 != 0 or year%400 ==0)):
            yrmo+= 1
    return (yrmo + day) * 86400   #24 * 60 * 60


def extendedntiles(varlist, ntiles=4, rootname="ntiles_", recodesample=None, varlabel=True, visible=False):
    """Calculate ntiles for a set of numeric variables and compute new variables 
    with the ntile values, possibly with a different sample definition.
    
    varlist is a sequence or simple, blank-separated string of one or more input variable names.
    ntiles is the number of ntiles and defaults to quartiles (4).
    For each variable in varlist a new variable is computed.  Its name consists of the rootname prefix followed by  the original name.
    The ntiles are calculated based on the sample in effect when this function is called.  If recodesample is specified, it defines the
    condition for the cases that will be recoded.  It could be a Boolean variable or an SPSS expression.
    The variables in cases that are excluded by the filter will have their old values or, for new variables, be sysmis.
    By default, the frequency tables underlying the ntile calculation are not displayed.  Set visible=True to display them.
    If varlabel is True, variable labels are constructed for the new variables consisting of the name of the input variable prefixed by
    "NTILE (ntile value):"
    If rootname is "", the input variables will be overwritten.
    
    Example:
    extendedntiles("x y z", 4, "ntiles_", sample="apply")
    would compute quartiles for each variable x, y, and z and produce new variables ntiles_x, ntiles_y, and ntiles_z.
    """
    
    stiles = str(ntiles)
    varlist = spssaux._buildvarlist(varlist)
    nvars = len(varlist)
    freqcmd = "FREQUENCIES " + " ".join(varlist) + "/FORMAT=NOTABLE/NTILES = " + stiles
    dsname, errlevel = spssaux.CreateDatasetOutput(freqcmd, omsid='Frequencies', subtype='Statistics', datasetvisible=visible)
    
    # The dataset produced has two rows to be discarded followed by a row for each ntile except the last (100%)
    # The fifth column identifies the percentile.  Each subsequent column contains the percentile cutpoint for the next variable.
    
    inputdsname = spssaux.GetActiveDatasetName()
    curs = spssdata.Spssdata(names=False, dataset=dsname)
    curs.fetchone()
    curs.fetchone()
    cutpointlist = []
    for i in range(nvars):
        cutpointlist.append([])
    if nvars == 1:
        baseinc = 1
    else:
        baseinc = 0
    for case in curs:
        for v in range(nvars):
            cutpointlist[v].append(case[v+5+baseinc])
    curs.close()
    if not visible:
        spss.Submit("DATASET CLOSE " + dsname)
    spss.Submit("DATASET ACTIVATE " + inputdsname)

    # create the new variables
    if recodesample:
        spss.Submit("DO IF " + recodesample)
        
    # generate RECODE and VARIABLE LABELS commands
    tailfragment = " (ELSE=" + str(ntiles)+")  INTO " + rootname
    for v in range(nvars):
        recodespec = " ".join(["(LOWEST THRU " + str(value) + "=" + str(i+1) + ")" for (i, value) in enumerate(cutpointlist[v])])
        spss.Submit("RECODE " + varlist[v] + recodespec + tailfragment + varlist[v])
        if varlabel:
            spss.Submit("VARIABLE LABEL " + rootname + varlist[v] + " 'NTILE(" + stiles + "): " + varlist[v]+"'")
    if recodesample:
        spss.Submit("END IF")
        
        
"""Collapse variable values below a certain count or frequency."""

def CollapseInfrequentValues(varlist, threshold=None, collapsevalue=None, rankcollapse=None, otherlabel="Other", weightvar=None, macronames=None, order="A"):
    """Collapse infrequent values in variables.
    
    varlist is a group of variable names as a blank-separated string or a list to be collapsed.  They must all be the same
    fundamental type (numeric or string).
    rankcollapse specifies collapsing based on rank.  By default, rankcollapse == None, collapsing is based on threshold.
    If rankcollapse is not None, it specifies the rank of the weighted count above which collapsing occurs.  For example, a value of 10 collapses
    all values ranked above 10.  If can be combined with threshold, in which case rank collapsing occurs first.  
    rankcollapse is truncated to an integer.
    threshold specifies the collapse criteria.  If >= 1, it specifies the (weighted) count below which the values will be collapsed.
    If it is < 1, it specifies the proportion of nonmissing cases below which values are collapsed.
    If it is <=0, no collapsing occurs, but this may still be useful in order to create the value macro.
    threshold and rankcollapse cannot be used together.
    Collapsing maps the infrequent values to a specified value and creates a value label for it.  The value labels of the values that
    are collapsed are removed.
    Collapsing occurs on the input variable.  To collapse on a new variable, duplicate the variable first using COMPUTE and
    APPLY DICTIONARY to copy the values and properties, or do this interactively using the Data Editor.
    collapsevalue specifies the value to be assigned to the collapsed cases, and otherlabel specifies the value label.
    if weightvar is specified, the case counted are weighted by this variable.
    If macronames is not None, it specifies a list of names corresponding in number to the varlist.  For each variable, a macro
    will be created listing the nonmissing values of the variable in order of frequency after any collapsing.
    order specifies whether the macro is in ascending or descending order.
    """
    
    if not order in ['A', 'D']:
        raise ValueError, "order must be 'A' or 'D'"
    if not [threshold,rankcollapse].count(None) == 1:
        raise ValueError, "Exactly one of threshold and rankcollapse values may be specified"
    if collapsevalue is None:
        raise ValueError, "A collapse value must be specified for CollapseInfrequentValues"
    varlist = spssaux._buildvarlist(varlist)
    nvars = len(varlist)
    vardict = spssaux.VariableDict(namelist=varlist)
    if weightvar and not weightvar in varlist:
        varlist.append(weightvar)
    vardicts = [dict() for i in range(len(varlist))]  #one dict for each variable being processed
    if not macronames is None:
        macronames = spssaux._buildvarlist(macronames)
    curs = spssdata.Spssdata(indexes=varlist, names=False, convertUserMissing=True)

    # accumulate distributions for all variables
    casetotal = 0.
    for case in curs:
        wt = weightvar and case[-1] or 1.0
        casetotal += wt
        for v in range(nvars):
            if not case[v] is None:   #ignore user and system missing values
                vardicts[v][case[v]] = vardicts[v].get(case[v], 0) + wt
    curs.CClose()
    vardict = spssaux.VariableDict(namelist=varlist)
    
    c = Collapser(vardict, threshold, collapsevalue, casetotal, rankcollapse, otherlabel, order)
    for v in range(nvars):
        if macronames is None:
            macarg = None
        else:
            macarg = macronames[v]
        c.collapse(vardict[v], vardicts[v], macarg)


class Collapser(object):
    """Variable value collapser based on empirical distribution"""
    
    def __init__(self, vardict, threshold, collapsevalue, casetotal, rankcollapse, otherlabel, order):
        """vardict is a VariableDict object for use in working with value labels
        threshold is the count or fraction for collapsing.
        collapsevalue is the value to use for collapsed values
        casetotal is the total number of cases
        If rankcollapse is not None, collapsing is based on rank; values with a rank higher than rankcollapse are collapsed.
        Threshhold is ignored in this case.
        otherlabel is the value label for the collapse value
        order is A or D for macro generation.
        weightvar is the weighting variable or None."""
        
        self.vardict = vardict
        self.collapsevalue = collapsevalue
        self.otherlabel = otherlabel
        self.order = order
        self.rankcollapse = rankcollapse
        
        if threshold >= 1:
            self.casethresh = threshold
        elif not threshold is None:
            self.casethresh = threshold * casetotal
        
    def collapse(self,var, vardist, macro):
        """collapse value(s) for variable according to empirical distribution.
        
        var is a Variable object
        vardist is a dictionary holding the empirical distribution
        macro if not None is the name of a macro to generate."""
        
        collapselist = []
        isstr = var.VariableType > 0
        macrovaluelist = []
        if not self.rankcollapse is None:
            collapselist, macrovaluelist = _rankcollapse(vardist, self.rankcollapse)
        else:
            for val, count in vardist.items():
                if count < self.casethresh:
                    collapselist.append(val)
                else:
                    macrovaluelist.append((count, str(val).rstrip()))
        if collapselist:
            if isstr:
                clist = [spssaux._smartquote(str(v)) for v in collapselist]
                cval = spssaux._smartquote(self.collapsevalue)
            else:
                clist = [str(v) for v in collapselist]
                cval = str(self.collapsevalue)
            cmd = "IF ANY(" + var.VariableName + "," + ", ".join(clist) + ") " +\
                var.VariableName + " = " + cval
            spss.Submit(cmd)
            valuelabels = self.vardict[var.VariableName].ValueLabels
            if not isstr:
                numerickeys = {}
                for k in valuelabels:
                    numerickeys[float(k)] = k
            if isstr:
                for val in collapselist:
                    val = val.rstrip()
                    if val in valuelabels:
                        del(valuelabels[val])
                valuelabels[self.collapsevalue.rstrip()] = self.otherlabel
            else:
                for val in collapselist:
                    fval = float(val)
                    if fval in numerickeys:
                        del(valuelabels[numerickeys[fval]])
                valuelabels[self.collapsevalue] = self.otherlabel
            self.vardict[var.VariableName].ValueLabels = valuelabels
        if macro is None:
            return
        macrovaluelist.sort(reverse= (self.order == "D"))
        if isstr:
            macrovaluelist = [spssaux._smartquote(val) for (count, val) in macrovaluelist]
        else:
            macrovaluelist = [str(val) for count, val in macrovaluelist]
        if collapselist:
            macrovaluelist.append(cval)
        spss.SetMacroValue(macro, " ".join(macrovaluelist))
            
def _rankcollapse(vardist, rankcollapse):
    """Return list of keys in vardist to collapse according to rankcollapse and macro valuelist.
    
    vardist is a dictionary of values and counts
    rankcollapse is the rank value above which to collapse"""
    
    collapselist = []
    macrovalues = []
    rankcollapse = int(rankcollapse)
    
    slis = sorted(vardist.items(), key=operator.itemgetter(1), reverse=True)
    for i in xrange(rankcollapse, len(slis)):
        collapselist.append(slis[i][0])
    for i in xrange(rankcollapse):
        macrovalues.append((slis[i][1], str(slis[i][0]).rstrip()))
    return collapselist, macrovalues
    