# state.py
# windows utility to create Spss KEEP or DROP statements.

import string, textwrap, os.path, inspect
import spss, spssaux

try:
    import win32clipboard as w
except ImportError:
    print "--> Note. Clipboard module not detected. Ctrl-V will not work"
    print "--> Install the PyWin32 package if you want to use the clipboard functionality"

__author__  = "Albert-Jan Roskam"
__version__ = "1.0.4"

def make_statement(sav=None, sps=None):
    """ Generate an SPSS keep statement for use in GET, ADD FILES, MATCH, SAVE, etc.
    For example: GET FILE = 'somefile.sav' /KEEP = this_function's_output.
    Only the variables that are actually used in syntax 'sps' and which are present
    in dataset 'sav' are kept for faster processing."""
    sav = check_data(sav)
    sps = check_syntax(sps)

    punct  = "()=,'\".<>/*"
    trans  = string.maketrans(punct, " " * len(punct))
    syntax = string.translate(sps, trans) # get rid of all the punctuation

    maybe_vars  = [element.lower().strip() for element in syntax.split()]
    all_vars    = [v.lower() for v in spssaux.GetVariableNamesList()]
    used_vars   = set(all_vars).intersection(set(maybe_vars))
    unused_vars = set(all_vars).difference(set(used_vars))

    if len(all_vars) == len(used_vars):
        statement = "/keep = all."
    elif len(unused_vars) < len(used_vars):
        statement = "/drop = %s." % " ".join(sorted(unused_vars))
    else:
        statement = "/keep = %s." % " ".join(sorted(used_vars))

    print_msg(statement, sav)
    try:
        copy_to_clipboard(statement, sav)
    except NameError:
        pass

def check_data(sav):
    """ Checks if and how Spss data are specified. Auxiliary function - not to be called directly."""
    try:
        if sav is None:
            sav = spssaux.getDatasetInfo()
            return sav
        elif os.path.isfile(sav) and sav.lower().endswith(".sav"):
            spss.Submit(r"get file = '%s'." % sav)
            return sav
    except TypeError, AttributeError:
        raise "--> Wrong SAV Input: %s" % sav.__class__

def check_syntax(sps=None):
    """ Checks if and how Spss syntax is specified. Auxiliary function - not to be called directly."""
    try:
        if sps is None:
            prompt = "--> Enter SPSS syntax: "
            sps = raw_input(prompt)
            return sps
        else: 
            return sps
    except EOFError:# raw_input can not be used in inline Python code in Spss
        try:  
            if os.path.isfile(sps) and sps.lower().endswith(".sps"):
                syntax = open(sps, "rb")
                sps = " ".join(syntax.readlines())
                return sps
        except TypeError, AttributeError:
            raise "--> Wrong SPS Input: %s" % sps.__class__

def print_msg(statement, sav):
    print "--> Here's your syntax:"
    msg = textwrap.fill(statement, width=70, initial_indent="  ", subsequent_indent="  ")
    print "get file = '" + sav + "'\n" + msg

def copy_to_clipboard(statement, sav):
    """ Copies KEEP statement to MS Windows clipboard. Auxiliary function - not to be called directly."""
    try:
        funcs = inspect.getmembers(w, inspect.isroutine)
        clipboard_module_is_present = [func for func in funcs if func[0] == "SetClipboardText"]
        if clipboard_module_is_present:
            print "--> Use CTRL-V to paste your syntax"
            msg = textwrap.fill(statement, width=70, initial_indent="  ", subsequent_indent="  ")
            set_text(mystring="get file = '" + sav + "'\r\n" + msg)
    except NameError: # w is not defined, i.e. clipboard module is not defined.
        pass

def set_text(mystring):
    """ Function that sets MS Windows clipboard text. Auxiliary function - not to be called directly."""
    w.OpenClipboard()
    w.EmptyClipboard()
    w.SetClipboardText(mystring)
    w.CloseClipboard()

if __name__ == "__main__":
    spss.Submit(r"get file = 'c:/program files/spss/employee data.sav'.") 
    make_statement()
##    make_statement(sav="c:/program files/spss/employee data.sav")
##    make_statement(sps=r""" crosstabs educ educat by salary""")
##    make_statement(sps=r""" gender bdate educ jobcat salary salbegin jobtime prevexp minority gender bdate educ jobcat""")
##    make_statement(sav="c:/program files/spss/employee data.sav", sps=r""" crosstabs educ educat by salary""") 
##    make_statement(sav=1.0, sps={}) ## yields error!


##BEGIN PROGRAM.
##import sys, spss
##sys.path.append(r"d:/temp")
##import state
##spss.Submit(r"get file='c:\program files\spss\Employee data.sav'.")
##state.make_statement(sps=r"""
##crosstabs educ educat by salary
##""")
##END PROGRAM.

##BEGIN PROGRAM.
##import state
##state.make_statement(sps=r"""
##crosstabs educ educat by salary
##""")
##END PROGRAM.
